// --------------------------------------------------------------------------------------
//
// FishFace40.CS                                                   Ulrich Mller 14.05.05
//                                                                 PCS1700   C# Final SP2
// Assembly FishFace40 mit Funktionen zum Betrieb der fishcertechnik Interfaces
//                     Basis umFish40.DLL / FtLib.LIB (Knobloch | fischertechnik)
// Klasse FishFace     : Kapselung und Erweiterung der Funktionen von umFish40.DLL
// Klasse FishRobot    : Methoden fr den Industry Robot & Co.
// Klasse FishStep     : Methoden fr Schrittmotoren
// Klasse FishFaceException : throw FishFace Ausnahmen
//
//
// --------------------------------------------------------------------------------------

using System;
using System.Runtime.InteropServices;
using System.Windows.Forms;

namespace FishFace40
{
	public enum IFTypen {
		ftROBO_first_USB       = 0,
		ftIntelligent_IF       = 10,
		ftIntelligent_IF_Slave = 20,
		ftROBO_IF_IIM          = 50,
		ftROBO_IF_USB          = 60,
		ftROBO_IF_COM          = 70,
		rfROBO_IF_Over_RF      = 80,
		ftROBO_IO_Extension    = 90,
		ftROBO_RF_Datalink     = 110
	}

	public enum Port {COM1 = 1, COM2, COM3, COM4}

	public enum Dir {
		Aus    = 0, Off   = Aus,
		Ein    = 1, On    = Ein,
		Links  = 1, Left  = Links,
		Rechts = 2, Right = Rechts
	}

	public enum Inp {
		AX = 1, AY, AXS1, AXS2, AXS3,
		A1 = 1, A2, AV, AZ,
		D1 = 1, D2,
		I1 = 1, I2, I3, I4, I5, I6, I7, I8, I9, I10, I11, I12, I13, I14, I15, I16,
		I17, I18, I19, I20, I21, I22, I23, I24, I25, I26, I27, I28, I29, I30, I31, I32
	}
	public enum IRCode {All = 0, Code1, Code2}
	public enum IRKeys {M3R = 1, M3L, M1, M2, M3, Code2, M1BW, M1FW, M2L, M2R, Code1
	}
	public enum Out {
		M1 = 1, M2, M3, M4, M5, M6, M7, M8, M9, M10, M11, M12, M13, M14, M15, M16,
		O1 = 1, O2, O3, O4, O5, O6, O7, O8, O9, O10, O11, O12, O13, O14, O15, O16,
		O17, O18, O19, O20, O21, O22, O23, O24, O25, O26, O27, O28, O29, O30, O31, O32
}

	public enum Speed {
		Off = 0, L1, L2, L3, Half, L5, L6, Full
	}
	public enum Wait {
		Ende = 0, Time, NotHalt, ESC
	}

	public struct DeviceData {
		public string Name;
		public int    Type;
		public int    SerialNr;
		public string Firmware;
	}

	public class FishFaceException: ApplicationException {
		public FishFaceException() {}
		public FishFaceException(string message) : base(message) {}
		public FishFaceException(string message, Exception inner) : base(message, inner){}
	}

	public class FishFace
	{
		// --- umFish40.DLL Deklarationen ----------------------------------------
		[DllImport("umFish40.DLL")]
		private static extern uint rbOpenInterfaceUSB(int ifTyp, int SerialNr);
		[DllImport("umFish40.DLL")]
		private static extern uint rbOpenInterfaceCOM(int ifTyp, 
			int ComNr, int AnalogZyklen);
		[DllImport("umFish40.DLL")]
		private static extern uint rbCloseInterface(uint iHandle);
		[DllImport("umFish40.DLL")]
		private static extern int rbGetInput(uint iHandle, int InputNr);
		[DllImport("umFish40.DLL")]
		private static extern int rbGetInputs(uint iHandle);
		[DllImport("umFish40.DLL")]
		private static extern int rbGetAnalog(uint iHandle, int AnalogNr);
		[DllImport("umFish40.DLL")]
		private static extern int rbGetIRKey(uint iHandle, int Code, int KeyNr);
		[DllImport("umFish40.DLL")]
		private static extern int rbGetVoltage(uint iHandle, int VoltNr);
		[DllImport("umFish40.DLL")]
		private static extern uint rbSetMotor(uint iHandle, int MotNr, int Dir);
		[DllImport("umFish40.DLL")]
		private static extern uint rbSetMotorEx(uint iHandle, 
			int MotNr, int Dir, int Speed);
		[DllImport("umFish40.DLL")]
		private static extern int rbGetMotors(uint iHandle);
		[DllImport("umFish40.DLL")]
		private static extern uint rbSetMotors(uint iHandle, int MotorStatus);
		[DllImport("umFish40.DLL")]
		private static extern uint rbSetMotorsEx(uint iHandle, int MotorStatus,
			int SpeedStatus, int SpeedStatus16);
		[DllImport("umFish40.DLL")]
		private static extern int rbGetModeStatus(uint iHandle, int MotNr);
		[DllImport("umFish40.DLL")]
		private static extern uint rbSetModeStatus(uint iHandle, int MotNr, int Mode);
		[DllImport("umFish40.DLL")]
		private static extern uint rbSetLamp(uint iHandle, int LampNr, int OnOff);
		[DllImport("umFish40.DLL")]
		private static extern uint rbSetLampEx(uint iHandle, 
			int LampNr, int OnOff, int Power);
		[DllImport("umFish40.DLL")]
		private static extern uint rbRobMotor(uint iHandle, 
			int MotNr, int Dir, int Speed, int ICount);
		[DllImport("umFish40.DLL")]
		private static extern uint rbRobMotors(uint iHandle, 
			int MotorStatus, uint SpeedStatus, int SpeedStatus16, int ModeStatus);
		[DllImport("umFish40.DLL")]
		private static extern int rbGetCounter(uint iHandle, int CounterNr);
		[DllImport("umFish40.DLL")]
		private static extern uint rbSetCounter(uint iHandle, int CounterNr, int ICount);
		[DllImport("umFish40.DLL")]
		private static extern uint rbClearCounters(uint iHandle);
		[DllImport("umFish40.DLL")]
		private static extern uint rbGetActDeviceType(uint iHandle);
		[DllImport("umFish40.DLL")]
		private static extern uint rbGetActDeviceSerialNr(uint iHandle);
		[DllImport("umFish40.DLL")]
		private static extern uint rbGetActDeviceFirmwareNr(uint iHandle);


		// --- Win API Deklarationen ----------------------------------------------------

		[DllImport("kernel32")]
		protected static extern void Sleep(uint mSek);
		[DllImport("kernel32")]
		protected static extern uint GetTickCount();
		[DllImport("user32")]
		protected static extern uint GetAsyncKeyState(uint vKey);
		protected const uint ESCAPE = 0x1B;

		// --- private (globale) Variable ----------------------------------------

		private uint   iHandle;        // --- Handle auf aktive umFish40-Instanz
		private bool   FOpen;          // --- OpenInterface erfolgreich
		private bool   FNotHalt;       // --- NotHalt
		private bool   FDoEvents;      // --- mit/ohne Application.DoEvents

		public const uint rbFehler = 0xE0000001;
		protected const uint rbPollInterval = 10;

		// --- FishFace : Constructor ---------------------------------------------

		public FishFace() {}

		// --- ffentliche Methoden und Eigenschaften ------------------------------
		
		// --- ClearCounter : Lschen eines Counters ------------------------------

		public void ClearCounter(int InputNr) {
			if(!FOpen) throw(new FishFaceException("KeinOpen.ClearCounter"));
			if(rbSetCounter(iHandle, InputNr, 0) == rbFehler)
				throw(new FishFaceException("InterfaceProblem.ClearCounter"));
		}

		// --- ClearCounters : Lschen aller Counter -----------------------------

		public void ClearCounters() {
			if(!FOpen) throw(new FishFaceException("KeinOpen.ClearCounters"));
			if(rbClearCounters(iHandle) == rbFehler)
				throw(new FishFaceException("InterfaceProblem.ClearCounters"));
		}

		// --- ClearMotors : Lschen aller M-Ausgnge ----------------------------

		public void ClearMotors() {
			if(!FOpen) throw(new FishFaceException("KeinOpen.ClearMotors"));
			if(rbSetMotors(iHandle, 0) == rbFehler)
				throw(new FishFaceException("InterfaceProblem.ClearMotors"));
		}
		// --- CloseInterface : Schlieen des Interface --------------------------

		public void CloseInterface() {
			if(FOpen) if(rbCloseInterface(iHandle) == rbFehler)
				throw(new FishFaceException("InterfaceProblem.CloseInterface"));
			FOpen = false;
		}

		// --- Finish : Feststellen Endewunsch (NotHalt, Escape, InputNr -----

		public bool Finish() {
			return Finish(0);
		}
		public bool Finish(Inp InputNr) {return Finish((int)InputNr);}
		public bool Finish(int InputNr) {
			if(FDoEvents) Application.DoEvents();
			if(!FOpen) throw(new FishFaceException("KeinOpen.Finish"));
			if(FNotHalt || GetAsyncKeyState(ESCAPE) !=0) return true;
			if((uint)GetInputs() == rbFehler) 
				throw(new FishFaceException("InterfaceProblem.Finish"));
			if((InputNr > 0) && GetInput(InputNr)) return true;
			return false;
		}
		public bool Finish(IRCode Code, IRKeys KeyNr) {
			return Finish((int)Code, (int)KeyNr);
		}
		public bool Finish(int Code, int KeyNr) {
			if(FDoEvents) Application.DoEvents();
			if(!FOpen) throw(new FishFaceException("KeinOpen.Finish"));
			if(FNotHalt || GetAsyncKeyState(ESCAPE) !=0) return true;
			if((uint)GetInputs() == rbFehler) 
				throw(new FishFaceException("InterfaceProblem.Finish"));
			if(GetIRKey(Code, KeyNr)) return true;
			return false;
		}

		// --- ActDevice : Auslesen der DeviceDaten ----------------------

		public DeviceData ActDevice {
			get {
				DeviceData dd;
				dd.Type     = (int)rbGetActDeviceType(iHandle);
				dd.SerialNr = (int)rbGetActDeviceSerialNr(iHandle);
				uint f  = rbGetActDeviceFirmwareNr(iHandle);
				dd.Firmware = String.Format("{0}.{1:00}.{2:00}.{3:00}", 
					(f & 0xFF000000) >> 24, (f & 0xFF0000) >> 16, 
					(f & 0xFF00) >> 8,      (f & 0xFF));
				dd.Name = Convert.ToString((IFTypen)dd.Type);
				return dd;
			}
		}

		// --- GetAnalog : Auslesen Analogwert -------------------------------

		public int GetAnalog(Inp AnalogNr) {return GetAnalog((int)AnalogNr);}
		public int GetAnalog(int AnalogNr) {
			if(FDoEvents) Application.DoEvents();
			if(!FOpen) throw(new FishFaceException("KeinOpen.GetAnalog"));
			int res = rbGetAnalog(iHandle, AnalogNr);
			if((uint)res == rbFehler) 
				throw(new FishFaceException("InterfaceProblem.GetAnalog"));
			return res;
		}

		// --- GetCounter : Auslesen eines ImpulsCounts ---------------------

		public int GetCounter(Inp InputNr) {return GetCounter((int)InputNr);}
		public int GetCounter(int InputNr) {
			if(!FOpen) throw(new FishFaceException("KeinOpen.GetCounter"));
			int res = rbGetCounter(iHandle, InputNr);
			if((uint)res == rbFehler)
				throw(new FishFaceException("InterfaceProblem.ClearCounter"));
			return res;
		}

		// --- GetInput : Abfragen des Status eines I-Einganges --------------

		public bool GetInput(Inp InputNr) {return GetInput((int)InputNr);}
		public bool GetInput(int InputNr) {
			if(FDoEvents) Application.DoEvents();
			if(!FOpen) throw(new FishFaceException("KeinOpen.GetInput"));
			int res = rbGetInput(iHandle, InputNr);
			if((uint)res == rbFehler)
				throw(new FishFaceException("InterfaceProblem.GetInput"));
			return (res == 1) ? true : false;
		}

		// --- GetInputs : Auslesen des Status aller I-Eingnge ---------------

		public int GetInputs() {
			if(!FOpen) throw(new FishFaceException("KeinOpen.GetInputs"));
			if(FDoEvents) Application.DoEvents();
			int res = rbGetInputs(iHandle);
			if((uint)res == rbFehler)
				throw(new FishFaceException("InterfaceProblem.GetInputs"));
			return res;
		}

		// --- GetIRKey : Auslesen Wert der IR Control --------------------------

		public bool GetIRKey(IRCode Code, IRKeys KeyNr) {
			return GetIRKey((int)Code, (int)KeyNr);}
		public bool GetIRKey(int Code, int KeyNr) {
			if(FDoEvents) Application.DoEvents();
			if(!FOpen) throw(new FishFaceException("KeinOpen.GetIRKey"));
			int res = rbGetIRKey(iHandle, Code, KeyNr);
			if((uint)res == rbFehler) 
				throw(new FishFaceException("InterfaceProblem.GetIRKey"));
			return (res == 1) ? true : false;
		}

		// --- GetVoltage : Auslesen Spannungswert -------------------------------

		public int GetVoltage(Inp VoltNr) {return GetVoltage((int)VoltNr);}
		public int GetVoltage(int VoltNr) {
			if(FDoEvents) Application.DoEvents();
			if(!FOpen) throw(new FishFaceException("KeinOpen.GetVoltage"));
			int res = rbGetVoltage(iHandle, VoltNr);
			if((uint)res == rbFehler) 
				throw(new FishFaceException("InterfaceProblem.GetVoltage"));
			return res;
		}

		// --- NotHalt : Abbruchwunsch -----------------------------------------

		public bool NotHalt {
			get{return FNotHalt;}
			set{FNotHalt = value;}
		}

		// --- OpenInterface : Verbindung zum Interface herstellen ------------

		public void OpenInterface(IFTypen ifTyp, int SerialNr) {
			OpenInterface(ifTyp, SerialNr, true);
		}
		public void OpenInterface(IFTypen ifTyp, int SerialNr, bool DoEvents) {
			iHandle = rbOpenInterfaceUSB((int)ifTyp, SerialNr);
			if((uint)iHandle == rbFehler)
				throw(new FishFaceException("InterfaceProblem.Open"));
			rbSetMotors(iHandle, 0);
			FNotHalt = false;
			FOpen = true;
			FDoEvents = DoEvents;
		}
		public void OpenInterface(IFTypen ifTyp, Port ComNr, int AnalogZyklen) {
			OpenInterface(ifTyp, ComNr, AnalogZyklen, true);
		}
		public void OpenInterface(IFTypen ifTyp,
			Port ComNr, int AnalogZyklen, bool DoEvents) {
			OpenInterface(ifTyp, (int)ComNr, AnalogZyklen, DoEvents);}
		public void OpenInterface(IFTypen ifTyp, int ComNr, int AnalogZyklen) {
			OpenInterface(ifTyp, ComNr, AnalogZyklen, true);
		}
		public void OpenInterface(IFTypen ifTyp, 
			int ComNr, int AnalogZyklen, bool DoEvents) {
			iHandle = rbOpenInterfaceCOM((int)ifTyp, ComNr, AnalogZyklen);
			if((uint)iHandle == rbFehler)
				throw(new FishFaceException("InterfaceProblem.Open"));
			rbSetMotors(iHandle, 0);
			FNotHalt = false;
			FOpen = true;
			FDoEvents = DoEvents;
		}

		// --- Outputs : Auslesen aller M-Ausgnge ---------------------------

		public int Outputs {
			get {
				if (!FOpen) throw(new FishFaceException("KeinOpen.Outputs"));
				int res = rbGetMotors(iHandle);
				if ((uint)res == rbFehler)
					throw(new FishFaceException("InterfaceProblem.Outputs"));
				return res;
			}
		}

		// --- Pause : Programm anhalten --------------------------------------
	
		public void Pause(int mSek) {
			if (!FOpen) throw(new FishFaceException("KeinOpen.Pause"));
			uint sZeit = GetTickCount();
			do {
				if(FDoEvents) Application.DoEvents();
				if ((uint)GetInputs() == rbFehler)
					throw(new FishFaceException("InterfaceProblem.Pause"));
				if (FNotHalt || GetAsyncKeyState(ESCAPE) !=0)return;
				Sleep(rbPollInterval);
			}	while (GetTickCount() - sZeit < mSek);
		}
		
		// --- SetCounter : Setzen ImpulsZhler ------------------------------

		public void SetCounter(Inp InputNr, int ICount) {
			SetCounter((int)InputNr, ICount);}
		public void SetCounter(int InputNr, int ICount) {
			if(!FOpen) throw(new FishFaceException("KeinOpen.SetCounter"));
			if(rbSetCounter(iHandle, InputNr, ICount) == rbFehler)
				throw(new FishFaceException("InterfaceProblem.SetCounter"));
		}

		// --- SetLamp : Setzen eines O-Ausganges.

		public void SetLamp(Out LampNr, Dir OnOff) {
			SetLamp((int)LampNr, (int)OnOff);}
		public void SetLamp(int LampNr, int OnOff) {
			SetLamp(LampNr, OnOff, 7);}
		public void SetLamp(Out LampNr, Dir OnOff, int Power) {
			SetLamp((int)LampNr, (int)OnOff, Power);}
		public void SetLamp(int LampNr, int OnOff, int Power) {
			if(FDoEvents) Application.DoEvents();
			if (!FOpen) throw(new FishFaceException("KeinOpen.SetLamp"));
			if(rbSetLampEx(iHandle, LampNr, OnOff, Power) == rbFehler)
				throw(new FishFaceException("InterfaceProblem.SetLamp"));
		}

		// --- SetMotor : Setzen eines M-Ausganges ---------------------------

		public void SetMotor(Out MotorNr, Dir Direction, Speed Speed, int ICount) {
			SetMotor((int)MotorNr, (int)Direction, (int)Speed, ICount);}
		public void SetMotor(int MotorNr, int Direction, int Speed, int ICount) {
			if(FDoEvents) Application.DoEvents();
			if (!FOpen) throw(new FishFaceException("KeinOpen.SetMotor"));
			if(rbRobMotor(iHandle, MotorNr, Direction, Speed, ICount) == rbFehler)
				throw(new FishFaceException("InterfaceProblem.SetMotor"));
		}
		public void SetMotor(Out MotorNr, Dir Direction, Speed Speed) {
			SetMotor((int)MotorNr, (int)Direction, (int)Speed);}
		public void SetMotor(int MotorNr, int Direction, int Speed) {
			if(FDoEvents) Application.DoEvents();
			if (!FOpen) throw(new FishFaceException("KeinOpen.SetMotor"));
			if(rbSetMotorEx(iHandle, MotorNr, Direction, Speed) == rbFehler)
				throw(new FishFaceException("InterfaceProblem.SetMotor"));
		}
		public void SetMotor(Out MotorNr, Dir Direction) {
			SetMotor((int)MotorNr, (int)Direction);}
		public void SetMotor(int MotorNr, int Direction) {
			if(FDoEvents) Application.DoEvents();
			if (!FOpen) throw(new FishFaceException("KeinOpen.SetMotor"));
			if(rbSetMotor(iHandle, MotorNr, Direction) == rbFehler)
				throw(new FishFaceException("InterfaceProblem.SetMotor"));
		}

		// --- SetMotors : Setzen aller M-Ausgnge ----------------------------

		public void SetMotors(int MotorStatus, 
			int SpeedStatus, int SpeedStatus16, int ModeStatus) {
			if(FDoEvents) Application.DoEvents();
			if (!FOpen) throw(new FishFaceException("KeinOpen.SetMotors"));
			if(rbRobMotors(iHandle, MotorStatus, 
				(uint)SpeedStatus, SpeedStatus16, ModeStatus) == rbFehler)
				throw(new FishFaceException("InterfaceProblem.SetMotors"));
		}
		public void SetMotors(int MotorStatus, 
			int SpeedStatus, int SpeedStatus16) {
			if(FDoEvents) Application.DoEvents();
			if (!FOpen) throw(new FishFaceException("KeinOpen.SetMotors"));
			if(rbSetMotorsEx(iHandle, MotorStatus, 
				SpeedStatus, SpeedStatus16) == rbFehler)
				throw(new FishFaceException("InterfaceProblem.SetMotors"));
		}
		public void SetMotors(int MotorStatus) {
			if(FDoEvents) Application.DoEvents();
			if (!FOpen) throw(new FishFaceException("KeinOpen.SetMotors"));
			if(rbSetMotors(iHandle, MotorStatus) == rbFehler)
				throw(new FishFaceException("InterfaceProblem.SetMotors"));
		}
		
		// --- Version --------------------------------------------------------------

		public static string Version {
			get{return "4.0.3.01";}
		}

		// --- WaitForChange : Warte auf NrOfChanges Impulse von InputNr -----------

		public void WaitForChange(Inp InputNr, int NrOfChanges) {
			WaitForChange((int)InputNr, NrOfChanges);}
		public void WaitForChange(int InputNr, int NrOfChanges) {
			WaitForChange(InputNr, NrOfChanges, 0);
		}
		public void WaitForChange(Inp InputNr, int NrOfChanges, Inp TermInputNr) {
			WaitForChange((int)InputNr, NrOfChanges, (int)TermInputNr);}
		public void WaitForChange(int InputNr, int NrOfChanges, int TermInputNr) {
			if (!FOpen) throw(new FishFaceException("KeinOpen.WaitForChange"));
			rbSetCounter(iHandle, InputNr, 0);

			while (rbGetCounter(iHandle, InputNr) < NrOfChanges) {		
				if(FNotHalt || GetAsyncKeyState(ESCAPE) !=0) return;
				if(TermInputNr>0) if(GetInput(TermInputNr)) return;
				if ((uint)GetInputs() == rbFehler)
					throw(new FishFaceException("InterfaceProblem.WaitForChange"));
				if(FDoEvents) Application.DoEvents();
			}
		}
		// --- WaitForHigh : Warten auf einen false/true-Durchgang von InputNr ------

		public void WaitForHigh(Inp InputNr) {WaitForHigh((int)InputNr);}
		public void WaitForHigh(int InputNr) {
			WaitForInput(InputNr, false);
			WaitForInput(InputNr, true);
		}

		// --- WaitForInput : Warten auf InputNr = OnOff ---------------------------

		public void WaitForInput(Inp InputNr) {WaitForInput((int)InputNr);}
		public void WaitForInput(int InputNr) {
			WaitForInput(InputNr, true);
		}
		public void WaitForInput(Inp InputNr, bool OnOff) {
			WaitForInput((int)InputNr, OnOff);}
		public void WaitForInput(int InputNr, bool OnOff) {
			if (!FOpen) throw(new FishFaceException("KeinOpen.WaitForInput"));
			do {
				if(FNotHalt || GetAsyncKeyState(ESCAPE) !=0) return;
				if ((uint)GetInputs() == rbFehler)
					throw(new FishFaceException("InterfaceProblem.WaitForInput"));
				Sleep(rbPollInterval);
				if(FDoEvents) Application.DoEvents();
			}   while(!GetInput(InputNr) == OnOff);
		}
		public void WaitForInput(IRCode Code, IRKeys KeyNr) {
			WaitForInput((int)Code, (int)KeyNr);}
		public void WaitForInput(int Code, int KeyNr) {
			WaitForInput(Code, KeyNr, true);
		}
		public void WaitForInput(IRCode Code, IRKeys KeyNr, bool OnOff) {
			WaitForInput((int)Code, (int)KeyNr, OnOff);}
		public void WaitForInput(int Code, int KeyNr, bool OnOff) {
			if (!FOpen) throw(new FishFaceException("KeinOpen.WaitForInput"));
			do {
				if(FNotHalt || GetAsyncKeyState(ESCAPE) !=0) return;
				if ((uint)GetInputs() == rbFehler)
					throw(new FishFaceException("InterfaceProblem.WaitForInput"));
				Sleep(rbPollInterval);
				if(FDoEvents) Application.DoEvents();
			}   while(!GetIRKey(Code, KeyNr) == OnOff);
		}

		// --- WaitForLow : Warten auf einen true/false-Durchgang von InputNr -----

		public void WaitForLow(Inp InputNr) {WaitForLow((int)InputNr);}
		public void WaitForLow(int InputNr) {
			WaitForInput(InputNr, true);
			WaitForInput(InputNr, false);
		}

		// --- WaitForMotors : Warten auf MotorReady-Ereignis oder Ablauf mSek -------

		public Wait WaitForMotors(int mSek, params int[] MotorNrs) {
			int c, i, j;
			uint t;

			if(!FOpen) throw(new FishFaceException("KeinOpen.WaitForMotors"));
			t = GetTickCount() + Convert.ToUInt32(mSek);
			do {
				c = 0;
				for(i=0; i<MotorNrs.Length; i++) {
					j = MotorNrs[i];
					c += rbGetModeStatus(iHandle, j);
				}

				if(c == 0) return Wait.Ende;

				// --- Abschalten der Motoren der Parameterliste ------
				if(FNotHalt || (GetAsyncKeyState(ESCAPE) != 0))
					for(i=0; i<MotorNrs.Length; i++) {
						j = (int)MotorNrs[i];
						rbSetMotor(iHandle, j, 0);
					};

				if(FNotHalt) return Wait.NotHalt;
				if(GetAsyncKeyState(ESCAPE) != 0)return Wait.ESC;
				if(FDoEvents) Application.DoEvents();
				Sleep(rbPollInterval);
				if ((uint)GetInputs() == rbFehler)
					throw(new FishFaceException("InterfaceProblem.WaitForMotors"));
			}   while((mSek == 0) || (t>GetTickCount()));
			return Wait.Time;
		}

		// --- WaitForPositionDown : Position = Counter - InputNr-Impulse ----------

		public void WaitForPositionDown(Inp InputNr, ref int Counter, int Position) {
			WaitForPositionDown((int)InputNr, ref Counter, Position);}
		public void WaitForPositionDown(int InputNr, ref int Counter, int Position) {
			WaitForPositionDown(InputNr, ref Counter, Position, 0);
		}
		public void WaitForPositionDown(Inp InputNr, ref int Counter, 
			int Position, Inp TermInputNr) {
			WaitForPositionDown((int)InputNr, ref Counter, Position, (int)TermInputNr);}
		public void WaitForPositionDown(int InputNr, ref int Counter, 
			int Position, int TermInputNr) {
			if (!FOpen) throw(new FishFaceException("KeinOpen.WaitForPositionDown"));
			WaitForChange(InputNr, Counter - Position, TermInputNr);
			Counter -= rbGetCounter(iHandle, InputNr);
		}

		// --- WaitForPositionUp : Position = Counter + InputNr-Impulse ------------

		public void WaitForPositionUp(Inp InputNr, ref int Counter, int Position) {
			WaitForPositionUp((int)InputNr, ref Counter, Position);}
		public void WaitForPositionUp(int InputNr, ref int Counter, int Position) {
			WaitForPositionUp(InputNr, ref Counter, Position, 0);
		}
		public void WaitForPositionUp(Inp InputNr, ref int Counter, 
			int Position, Inp TermInputNr) {
			WaitForPositionUp((int)InputNr, ref Counter, Position, (int)TermInputNr);}
		public void WaitForPositionUp(int InputNr, ref int Counter, 
			int Position, int TermInputNr) {
			if (!FOpen) throw(new FishFaceException("KeinOpen.WaitForPositionUp"));
			WaitForChange(InputNr, Position - Counter, TermInputNr);
			Counter += rbGetCounter(iHandle, InputNr);
		}

		// --- WaitForTime : Program mSek anhalten ----------------------------------

		public void WaitForTime(int mSek) {
			if (!FOpen) throw(new FishFaceException("KeinOpen.WaitForTime"));
			Pause(mSek);
		}
	}
	// -----------------------------------------------------------------------------------
	//
	// Klasse FishRobot : Betrieb von RobMotoren
	//
	// -----------------------------------------------------------------------------------

	public class FishRobot : FishFace {

		// --- globale Daten ------------------------------------------------
		public struct MotWerte {
			public int Nr;
			public int maxPos;
			public int actPos;
		}
		public MotWerte[] MotCntl;
		private int MotAnz;

		// --- Ereignisse ----------------------------------------

		public delegate void CommonDelegate(object sender, int[] ActPositions);
		public event CommonDelegate PositionChange;

		// --- FishRobot Konstruktor ------------------------------------

		public FishRobot(int[,] MotList) : base(){MotInit(MotList);}

		private void MotInit(int[,] MotList) {
			MotAnz =MotList.GetLength(0);
			MotCntl = new MotWerte[MotAnz];
			for(int i = 0; i < MotAnz; i++) {
				MotCntl[i].Nr     = MotList[i,0];
				MotCntl[i].maxPos = MotList[i,1];
				MotCntl[i].actPos = 0;
			}
		}

		// --- MoveHome : Simultanes Anfahren der Home-Positon ---------

		public void MoveHome() {
			int[] Mots = new int[MotAnz];
			for(int i = 0; i < MotAnz; i++) Mots[i] = MotCntl[i].Nr;
			try {
				foreach(int m in Mots) SetMotor(m, (int)Dir.Links, (int)Speed.Full, 9999);
				WaitForMotors(0, Mots);
				for(int i = 0; i < MotAnz; i++) MotCntl[i].actPos = 0;
			}
			catch(FishFaceException){
				throw(new FishFaceException("InterfaceProblem.MoveHome"));
			}
		}

		// --- MoveTo : Simultanes Anfahren einer vorgegebenen Position -----

		public void MoveTo(params int[] PosList){

			// --- Starten der Motoren ------------
			int anzPos = PosList.Length;
			try {
				for(int i = 0; i < anzPos; i++){
					int ZielPos = PosList[i];
					if(ZielPos > MotCntl[i].maxPos)ZielPos = MotCntl[i].maxPos;
					if(ZielPos > MotCntl[i].actPos) 
						SetMotor(MotCntl[i].Nr, (int)Dir.Rechts, (int)Speed.Full,
							ZielPos-MotCntl[i].actPos);
					else if (ZielPos < MotCntl[i].actPos) {
						SetMotor(MotCntl[i].Nr, (int)Dir.Links, (int)Speed.Full, 
							MotCntl[i].actPos-ZielPos);
					}
				}
				// --- Warten auf MotorReady ----
				int[] Mots = new int[anzPos];
				for(int i = 0; i < anzPos; i++) Mots[i] = MotCntl[i].Nr;
				int res = (int)Wait.Time;
				
				// --- Ausgeben der aktuellen Position ----
				while (res == (int)Wait.Time){
					if(PositionChange != null){
						int[] actPos = new int[anzPos];
						for(int i = 0; i < anzPos; i++) {
							int iNr = (Mots[i]-1)*2+2;
							int ZielPos = PosList[i];
							if(ZielPos > MotCntl[i].actPos)
								actPos[i] = ZielPos - GetCounter(iNr);
							else if(ZielPos < MotCntl[i].actPos) 
								actPos[i] = ZielPos + GetCounter(iNr);
							else actPos[i] = MotCntl[i].actPos;
						}
						PositionChange(this, actPos);
					}
					res = (int)WaitForMotors((int)(8*rbPollInterval), Mots);
				}
				// --- Aktualisieren MotCntl -------
				for(int i = 0; i < anzPos; i++){
					int iNr = (Mots[i]-1)*2+2;
					int ZielPos = PosList[i];
					if(ZielPos > MotCntl[i].actPos) 
						MotCntl[i].actPos = ZielPos - GetCounter(iNr);
					else if(ZielPos < MotCntl[i].actPos) 
						MotCntl[i].actPos = ZielPos + GetCounter(iNr);
					else MotCntl[i].actPos = ZielPos;
				}
				// --- Abschlieende Positionsausgabe --------
				if(PositionChange != null){
					int[] actPos = new int[anzPos];
					for(int i = 0; i < anzPos; i++) actPos[i] = MotCntl[i].actPos;
					PositionChange(this, actPos);					
				}
			}
			catch(FishFaceException){
				throw(new FishFaceException("InterfaceProblem.MoveTo/MoveDelta"));
			}
		}
		public void MoveDelta(params int[] DeltaList) {
			int anzPos = DeltaList.Length;
			int[] PosList = new int[anzPos];
			for(int i = 0; i < anzPos; i++)
				PosList[i] = MotCntl[i].actPos + DeltaList[i];
			MoveTo(PosList);
		}
	}
	// -----------------------------------------------------------------------------------
	// 
	// Klasse FishStep : Betrieb von Schrittmotoren, einzeln und im XY-Verbund
	//
	// -----------------------------------------------------------------------------------

	public class FishStep : FishFace {
		// --- globale Daten ----------------------------------------
		public struct MotWerte {
			public int maxPos;
			public int actPos;
			public bool outPos;
		}
		public MotWerte[] MotCntl;
		private int MotAnz;

		private enum MotS {MotLLL, MotLLR, MotLRL, MotLRR, MotRLL, MotRLR, MotRRL, MotRRR}
		private int[] MotMaske = {0xFFC0, 0xFF03, 0xFF0F, 0xF03F, 0xC0FF};
		private int[] BefMaske = {0x15, 0x25, 0x19, 0x29, 0x16, 0x26, 0x1A, 0x2A};
		private int[] MotShift = {0, 2, 4, 8, 16};

		// ------- Inline Funktionen ----------------------------------
		private int  EndNrX(int MotNr)  {return (MotNr-1)*2+1;}
		private int  EndNrY(int MotNr)  {return (MotNr+1)*2+1;}
		private int  Xmax(int MotNr)    {return MotCntl[MotNr-1].maxPos;}
		private int  Ymax(int MotNr)    {return MotCntl[MotNr+1].maxPos;}
		private int  Xact(int MotNr)    {return MotCntl[MotNr-1].actPos;}
		private int  Yact(int MotNr)    {return MotCntl[MotNr+1].actPos;}
		private bool Xout(int MotNr)    {return MotCntl[MotNr-1].outPos;}
		private bool Yout(int MotNr)    {return MotCntl[MotNr+1].outPos;}
		private int  XactInc(int MotNr) {return MotCntl[MotNr-1].actPos++;}
		private int  XactDec(int MotNr) {return MotCntl[MotNr-1].actPos--;}
		private int  YactInc(int MotNr) {return MotCntl[MotNr+1].actPos++;}
		private int  YactDec(int MotNr) {return MotCntl[MotNr+1].actPos--;}

		// --- Ereignisse -------------------------------------------
		public delegate void StepDelegate(object sender, int MotNr, int ActPosition);
		public event StepDelegate StepChange;
		public delegate void PlotDelegate(object sender, int MotNr, int Xact, int Yact);
		public event PlotDelegate PlotChange;

		// --- FishStep Konstruktor ----------------------------------
		public FishStep(int[,] MotList) : base() {
			MotAnz = MotList.GetLength(0);
			MotCntl = new MotWerte[MotAnz*2];
			for(int i = 0; i < MotAnz; i++) MotCntl[MotList[i,0]-1].maxPos = MotList[i,1];
		}
		
		// ---- interne Routinen Step ---------------------------------------

		// --- RechtsDreh : Fahren eines Rechtszyklus -------------------
		private void RechtsDreh(int MotNr) {
			SetMotor(MotNr, (int)Dir.Links);
			Pause((int)rbPollInterval);
			SetMotor(MotNr+1, (int)Dir.Links);
			Pause((int)rbPollInterval);
			SetMotor(MotNr, (int)Dir.Rechts);
			Pause((int)rbPollInterval);
			SetMotor(MotNr+1, (int)Dir.Rechts);
			Pause((int)rbPollInterval);
		}

		// --- LinksDreh : Fahren eines Linkszyklus -----------------
		private void LinksDreh(int MotNr) {
			SetMotor(MotNr, (int)Dir.Rechts);
			Pause((int)rbPollInterval);
			SetMotor(MotNr+1, (int)Dir.Links);
			Pause((int)rbPollInterval);
			SetMotor(MotNr, (int)Dir.Links);
			Pause((int)rbPollInterval);
			SetMotor(MotNr+1, (int)Dir.Rechts);
			Pause((int)rbPollInterval);
		}

		// --- Step-Methoden -----------------------------------------

		// --- StepHome : Fahren eines Motors auf Home-Position ------------
		public void StepHome(Out MotNr) {
			StepHome((int)MotNr);
		}
		public void StepHome(int MotNr) {
			try {
				while(!GetInput(EndNrX(MotNr))) {
					if(GetAsyncKeyState(ESCAPE) != 0) return;
					LinksDreh(MotNr);
				}
				MotCntl[MotNr-1].actPos = 0;
				MotCntl[MotNr-1].outPos = false;
			}
			catch(FishFaceException) { 
			  throw(new FishFaceException("InterfaceProblem.MoveHome"));}
		}
		
		// --- StepTo : Fahren Motor auf Zielposition -----
		public void StepTo(Out  MotNr, int Xabs) {
			StepTo((int)MotNr, Xabs);
		}
		public void StepTo(int MotNr, int Xabs) {
			try {
				int PosCount = 0;
				if(Xabs > Xact(MotNr)) {
					while(Xact(MotNr) < Xabs) {
						RechtsDreh(MotNr);
						XactInc(MotNr);
						if(Xabs >= MotCntl[MotNr-1].maxPos) return;
						if(GetAsyncKeyState(ESCAPE) !=0)    return;
						PosCount++;
						if(StepChange != null && PosCount == 10) {
							PosCount = 0;
							StepChange(this, MotNr, Xact(MotNr));
						}
					}
				}
				else {
					while(Xact(MotNr) > Xabs) {
						LinksDreh(MotNr);
						XactDec(MotNr);
						if(GetInput(EndNrX(MotNr)))      return;
						if(GetAsyncKeyState(ESCAPE) !=0) return;
						PosCount++;
						if(StepChange != null && PosCount == 10) {
							PosCount = 0;
							StepChange(this, MotNr, Xact(MotNr));
						}
					}
				}
				if(StepChange != null) StepChange(this, MotNr, Xact(MotNr));
			}
			catch(FishFaceException) {
			 throw(new FishFaceException("InterfaceProblem.StepTo"));}
		}

		// --- StepDelta : Fahren Motor um Xrel-Zyklen ---------------
		public void StepDelta(Out  MotNr, int Xrel) {
			StepDelta((int)MotNr, Xrel);
		}
		public void StepDelta(int MotNr, int Xrel) {
			try {
				StepTo(MotNr, Xact(MotNr)+Xrel);
			}
			catch(FishFaceException) {
			 throw(new FishFaceException("InterfaceProblem.StepDelta"));}
		}

		// --- Interne Routinen Plot ------
		// --- Zyklen in die acht Richtungen -----

		// --- BereichsLimit : Prfen, ob actKoordinaten im Plotbereich liegen ---
		private void BereichsLimit(int MotNr) {
			if(Xout(MotNr) || Yout(MotNr)) {
				MotCntl[MotNr-1].outPos = Xact(MotNr) < 0 || Xact(MotNr) > Xmax(MotNr);
				MotCntl[MotNr+1].outPos = Yact(MotNr) < 0 || Yact(MotNr) > Ymax(MotNr);
				if(Xout(MotNr) || Yout(MotNr)) return;
			}
			else {
				MotCntl[MotNr-1].outPos = Xact(MotNr) < 0 || Xact(MotNr) > Xmax(MotNr);
				MotCntl[MotNr+1].outPos = Yact(MotNr) < 0 || Yact(MotNr) > Ymax(MotNr);
			}
		}
		// --- Endtaster : Abfrage Endtaster im laufenden Betrieb ---
		private bool Endtaster(int MotNr) {
			if((GetInput(EndNrX(MotNr)) && (Xact(MotNr) != 0)) || 
				(GetInput(EndNrY(MotNr)) && (Yact(MotNr) != 0))) return true; 
			else return false;
		}

		// --- Zyklus : Abarbeiten des vorgegebenen Befehlszyklus ---
		private void Zyklus(int MotNr, params MotS[] Befehle){
			foreach(MotS mot in Befehle) {
				int MotStatus = Outputs & MotMaske[MotNr-1];
				SetMotors(MotStatus | BefMaske[(int)mot] << MotShift[MotNr-1]);
				Pause((int)rbPollInterval);
			}
		}

		// --- XMinusSchritt : Fahren in -X Richtung ---
		private void XMinusSchritt(int MotNr) {
			XactDec(MotNr);
			BereichsLimit(MotNr);
			if(Xout(MotNr)) return;
			Zyklus(MotNr, MotS.MotRLR, MotS.MotLLR, MotS.MotLRR, MotS.MotRRR);
		}

		// --- XPlusSchritt : Fahren in +X Richtung ---
		private void XPlusSchritt(int MotNr) {
			XactInc(MotNr);
			BereichsLimit(MotNr);
			if(Xout(MotNr)) return;
			Zyklus(MotNr, MotS.MotLRR, MotS.MotLLR, MotS.MotRLR, MotS.MotRRR);
		}

		// --- XpYpDiagonale : Fahren in +X/+Y Richtung ------
		private void XpYpDiagonale(int MotNr) {
			XactInc(MotNr);
			YactInc(MotNr);
			BereichsLimit(MotNr);
			if(Xout(MotNr)) {YPlusSchritt(MotNr); return;}
			if(Yout(MotNr)) {XPlusSchritt(MotNr); return;}
			Zyklus(MotNr, MotS.MotLRR, MotS.MotLLL, MotS.MotRLL, MotS.MotRRR);
		}

		// --- XpYmDiagonale : Fahren in +X/-Y Richtung ------
		private void XpYmDiagonale(int MotNr) {
			XactInc(MotNr);
			YactDec(MotNr);
			BereichsLimit(MotNr);
			if(Xout(MotNr)) {YMinusSchritt(MotNr); return;}
			if(Yout(MotNr)) {XPlusSchritt(MotNr); return;}
			Zyklus(MotNr, MotS.MotRRL, MotS.MotLRL, MotS.MotLLR, MotS.MotRLR, MotS.MotRRR);
		}
		// --- XmYmDiagonale : Fahren in -X/-Y Richtung ------
		private void XmYmDiagonale(int MotNr) {
			XactDec(MotNr);
			YactDec(MotNr);
			BereichsLimit(MotNr);
			if(Xout(MotNr)) {YMinusSchritt(MotNr); return;}
			if(Yout(MotNr)) {XMinusSchritt(MotNr); return;}
			Zyklus(MotNr, MotS.MotRLL, MotS.MotLLL, MotS.MotLRR, MotS.MotRRR);
		}

		// --- XmYpDiagonale : Fahren in -X/+Y Richtung ------
		private void XmYpDiagonale(int MotNr) {
			XactDec(MotNr);
			YactInc(MotNr);
			BereichsLimit(MotNr);
			if(Xout(MotNr)) {YPlusSchritt(MotNr); return;}
			if(Yout(MotNr)) {XMinusSchritt(MotNr); return;}
			Zyklus(MotNr, MotS.MotRLR, MotS.MotLLR, MotS.MotLRL, MotS.MotRRL, MotS.MotRRR);
		}

		// --- YMinusSchritt : Fahren in -Y Richtung -----
		private void YMinusSchritt(int MotNr) {
			YactDec(MotNr);
			BereichsLimit(MotNr);
			if(Yout(MotNr)) return;
			Zyklus(MotNr, MotS.MotRRL, MotS.MotLRL, MotS.MotLRR, MotS.MotRRR);
		}

		// --- YPlusSchritt : Fahren in +Y Richtung -----
		private void YPlusSchritt(int MotNr) {
			YactInc(MotNr);
			BereichsLimit(MotNr);
			if(Yout(MotNr)) return;
			Zyklus(MotNr, MotS.MotLRR, MotS.MotLRL, MotS.MotRRL, MotS.MotRRR);
		}

		// --- Plot-Methoden --------------
		
		// --- PlotHome : Fahren Motoren auf Home-Position(native : linke untere Ecke)
		public void  PlotHome(Out  MotNr) {
			PlotHome((int)MotNr);
		}
		public void  PlotHome(int MotNr) {
			try {
				MotCntl[MotNr-1].outPos = false;
				MotCntl[MotNr+1].outPos = false;
				ClearMotors();
				// --- Beide Motoren gegen Endtaster ----
				while (!(GetInput(EndNrX(MotNr)) || GetInput(EndNrY(MotNr)))) {
					if(GetAsyncKeyState(ESCAPE) != 0) return;
					Zyklus(MotNr, MotS.MotRLL, MotS.MotLLL, MotS.MotLRR, MotS.MotRRR);
				}
				// --- Motor X gegen Endtaster -----------
				while(GetInput(EndNrX(MotNr)) == false) {
					if(GetAsyncKeyState(ESCAPE) != 0) return;
					Zyklus(MotNr, MotS.MotRLR, MotS.MotLLR, MotS.MotLRR, MotS.MotRRR);
				}
				// --- Motor X frei von Endtaster -------
				while(GetInput(EndNrX(MotNr))) {
					Zyklus(MotNr, MotS.MotLRR, MotS.MotLLR, MotS.MotRLR, MotS.MotRRR);
				}
				// --- Motor Y gegen Endtaster -----------------
				while(GetInput(EndNrY(MotNr)) == false) {
					if(GetAsyncKeyState(ESCAPE) != 0) return;
					Zyklus(MotNr, MotS.MotRRL, MotS.MotLRL, MotS.MotLRR, MotS.MotRRR);
				}
				// --- Motor Y frei von Endtaster ----
				while(GetInput(EndNrY(MotNr))) {
					Zyklus(MotNr, MotS.MotLRR, MotS.MotLRL, MotS.MotRRL, MotS.MotRRR);
				}
				// --- Motoren zwei Schritt Abstand vom Ende --------
				XpYpDiagonale(MotNr);
				XpYpDiagonale(MotNr);
				MotCntl[MotNr-1].actPos = 0;
				MotCntl[MotNr+1].actPos = 0;
			}
			catch(FishFaceException) {
			 throw(new FishFaceException("InterfaceProblem.PlotHome"));}
		}
		
		// --- PlotDelta : Fahren Motoren um Xrel/Yrel-Werte(Plotzyklen ab links unten) ---
		public void PlotDelta(Out  MotNr, int Xrel, int Yrel) {
			PlotDelta((int)MotNr, Xrel, Yrel);
		}
		public void PlotDelta(int MotNr, int Xrel, int Yrel) {
			try {
				int Xdelta   = Math.Abs(Xrel);    // --- Anzahl Plotzyklen
				int Ydelta   = Math.Abs(Yrel);
				int Xnr      = 0;                 // --- Zyklenzhler
				int Ynr      = 0;
				int Delta    = 0;                 // --- Abstand zur optimalen Geraden
				int PosCount = 0;                 // --- Zhler fr PosAusgabe

				while(Xnr != Xdelta || Ynr != Ydelta) {
					if(GetAsyncKeyState(ESCAPE) != 0) return;
					if(Endtaster(MotNr))              return;
					if(Math.Abs(Delta+Ydelta) < Math.Abs(Delta+Ydelta-Xdelta)) {
						if(Xrel > 0) XPlusSchritt(MotNr); else XMinusSchritt(MotNr);
						Xnr++;
						Delta += Ydelta;
					}
					else {
						if(Math.Abs(Delta-Xdelta) < Math.Abs(Delta+Ydelta-Xdelta)) {
							if(Yrel > 0) YPlusSchritt(MotNr); else YMinusSchritt(MotNr);
							Ynr++;
							Delta -= Xdelta;
						}
						else {
							if(Xrel > 0) 
								if(Yrel > 0) XpYpDiagonale(MotNr); 
								else XpYmDiagonale(MotNr);
							else if(Yrel > 0) XmYpDiagonale(MotNr); 
							else XmYmDiagonale(MotNr);
							Xnr++;
							Ynr++;
							Delta += Ydelta - Xdelta;
						}
					}
					PosCount++;
					if(PlotChange != null && PosCount == 10) {
						PosCount = 0;
						PlotChange(this, MotNr, Xact(MotNr), Yact(MotNr));
					}
				}
				if(PlotChange != null) PlotChange(this, MotNr, Xact(MotNr), Yact(MotNr));
			}
			catch(FishFaceException) {
			 throw(new FishFaceException("InterfaceProblem.PlotDelta"));}
		}
				
		// --- PlotTo : Fahren Motoren auf Xabs/Yabs Position(Plotzyklen ab links unten)
		public void PlotTo(Out  MotNr, int Xabs, int Yabs) {
			PlotTo((int)MotNr, Xabs, Yabs);
		}
		public void PlotTo(int MotNr, int Xabs, int Yabs) {
			try {
				PlotDelta(MotNr, Xabs-Xact(MotNr), Yabs-Yact(MotNr));
			}
			catch(FishFaceException) {
			 throw(new FishFaceException("InterfaceProblem.PlotTo"));}
		}
	}
}
